import click
import rich.console
import requests

import os
import sys
import urllib3
import warnings
import signal
import types

from __version__ import __version__
import analysis
import visualization
import targets
import validation
import verbose
import utils
import print


warnings.simplefilter('ignore', urllib3.exceptions.InsecureRequestWarning)
CONTEXT_SETTINGS = dict(max_content_width=120, help_option_names=['-help'])


def ctrl_c_signal_handler(sig: int, frame: types.FrameType | None) -> None:
    """Handle the case where the user sends a CTRL-C keyboard interrupt.

    Args:
        sig (int): Signal number (e.g., SIGINT).
        frame (FrameType | None): Current stack frame at the time of the signal.
    """
    verbose.info("'Ctrl+C!' was pressed. Exit.", False)
    os._exit(1)


@click.command(
    context_settings=CONTEXT_SETTINGS,
    cls=utils.CustomCommand,
)
@click.version_option(
    __version__,
    '-version',
    cls=utils.CustomOption,
    category='DEBUG',
)
@click.option(
    '-no-color',
    help='Disable colors in CLI output.',
    is_flag=True,
    cls=utils.CustomOption,
    category='DEBUG',
)
@click.option(
    '-silent',
    help='Display only results in output.',
    is_flag=True,
    cls=utils.CustomOption,
    category='DEBUG',
)
@click.option(
    '-debug',
    help='Display debug messages.',
    is_flag=True,
    cls=utils.CustomOption,
    category='DEBUG',
)
@click.option(
    '-simulate',
    help='Display the parsed targets.',
    is_flag=True,
    cls=utils.CustomOption,
    category='DEBUG',
)
@click.option(
    '-target',
    help='Targets to analyze (comma-separated).',
    type=str,
    default='',
    cls=utils.CustomOption,
    category='INPUT',
)
@click.option(
    '-list',
    help='List of targets to analyze (file).',
    type=str,
    default='',
    callback=validation.validate_file_exists,
    cls=utils.CustomOption,
    category='INPUT',
)
@click.option(
    '-exclude-targets',
    help='Targets to exclude from analysis (comma-separated).',
    type=str,
    default='',
    cls=utils.CustomOption,
    category='INPUT',
)
@click.option(
    '-exclude-file',
    help='List of targets to exclude from analysis (file).',
    type=str,
    default='',
    callback=validation.validate_file_exists,
    cls=utils.CustomOption,
    category='INPUT',
)
@click.option(
    '-json',
    help='Write output in JSON lines format.',
    is_flag=True,
    cls=utils.CustomOption,
    category='OUTPUT',
)
@click.option(
    '-table',
    help='Write output in Table format.',
    is_flag=True,
    cls=utils.CustomOption,
    category='OUTPUT',
)
@click.option(
    '-visualize',
    help='Visualize output as a network graph image. Specify the filename.',
    cls=utils.CustomOption,
    callback=validation.validate_png_filename,
    category='OUTPUT',
)
@click.option(
    '-threads',
    help='The max number of worker threads.',
    type=int,
    default=10,
    cls=utils.CustomOption,
    category='TWEAK',
)
def cli(
    no_color: bool,
    silent: bool,
    debug: bool,
    simulate: bool,
    target: str,
    list: str,
    exclude_targets: str,
    exclude_file: str,
    json: bool,
    table: bool,
    visualize: str,
    threads: int,
) -> None:
    #
    # Global
    #

    verbose.SILENT = silent
    verbose.HIGHLIGHT = False
    verbose.SOFT_WRAP = True
    verbose.DEBUG = debug
    verbose.CONSOLE = rich.console.Console(no_color=no_color)

    #
    # CLI Signals
    #

    signal.signal(signal.SIGINT, ctrl_c_signal_handler)

    #
    # CLI Validation
    #

    if json and table:
        raise click.UsageError("You can not use '-json' and '-table' options at the same time.")

    #
    # Welcome
    #

    verbose.print_banner(silent)
    verbose.warning('Use with caution. You are responsible for your actions.')

    #
    # GeoIP2
    #

    verbose.info("Make sure 'geoip2-ipv4.csv' is downloaded.")
    GEOIP_SHA256 = '4d5b63c8a4dc7d78d395de2106f1ff1a38a654da67cdecde80ba7fe55db4cc7a'
    exe_location = os.path.dirname(os.path.abspath(__file__))
    geoip_filepath = os.path.join(exe_location, 'geoip2-ipv4.csv')
    if not validation._file_exists(geoip_filepath) or not validation._verify_sha256(geoip_filepath, GEOIP_SHA256):
        verbose.info('Download the geoip database from GitHub.')
        GEOIP_URL = 'https://raw.githubusercontent.com/datasets/geoip2-ipv4/refs/heads/main/data/geoip2-ipv4.csv'
        response = requests.get(GEOIP_URL, verify=False, timeout=3600)
        response.raise_for_status()
        with open(geoip_filepath, 'wb') as f:
            f.write(response.content)

    #
    # Input
    #

    targeter = targets.Targeter()
    if target != '':
        verbose.info("Parse targets from the 'target' CLI parameter.")
        targeter.parse_targets_str(target)
    elif list != '':
        verbose.info(f"Parse targets from the file located at '{list}'.")
        targeter.parse_targets_file(list)
    elif exclude_targets != '':
        verbose.info("Exclude targets from the 'exclude_targets' CLI parameter.")
        targeter.parse_exclusions_str(exclude_targets)
    elif exclude_file != '':
        verbose.info(f"Exclude targets from the file located at '{exclude_file}'.")
        targeter.parse_exclusions_file(exclude_file)
    elif not sys.stdin.isatty():
        verbose.info('Parse targets from the STDIN.')
        targeter.parse_targets_file('-')

    #
    # No Targets
    #

    if targeter.total_count() == 0:
        raise click.UsageError('You must supply at least one target.')

    #
    # Simulation
    #

    if simulate:
        verbose.info('Simulate and print the parsed targets.')
        targeter.print_targets()
        exit(1)

    #
    # Analysis
    #

    analyzer = analysis.Analyzer()
    analyzer.parse_geoip_data(geoip_filepath)
    verbose.info('Analyze the targets.')
    if len(targeter.ipv4s) > 0:
        analyzer.analyze_ipv4s(targeter.ipv4s, threads)
    if len(targeter.cidrs_v4) > 0:
        analyzer.analyze_cidrs(targeter.cidrs_v4, threads)
    if len(targeter.fqdns) > 0:
        analyzer.analyze_fqdns(targeter.fqdns, threads)
    if len(targeter.urls) > 0:
        analyzer.analyze_urls(targeter.urls, threads)

    #
    # stdout
    #

    if visualize:
        verbose.info('Visualize the targets as a network graph.')
        visualizer = visualization.Visualizer()
        visualizer.create_visualization_image(
            analyzer.analyzed_ipv4s,
            analyzer.analyzed_cidrs,
            analyzer.analyzed_fqdns,
            analyzer.analyzed_urls,
            visualize,
        )
        return

    verbose.info('Print the results in the stdout.')
    verbose.SILENT = False
    if len(targeter.ipv4s) > 0:
        if table:
            print.Printer.print_ipv4s_as_table(analyzer.analyzed_ipv4s)
        elif json:
            print.Printer.print_as_json(analyzer.analyzed_ipv4s)
        else:
            print.Printer.print_ipv4s_as_table(analyzer.analyzed_ipv4s)
    if len(targeter.cidrs_v4) > 0:
        if table:
            print.Printer.print_cidrs_as_table(analyzer.analyzed_cidrs)
        elif json:
            print.Printer.print_as_json(analyzer.analyzed_cidrs)
        else:
            print.Printer.print_cidrs_as_table(analyzer.analyzed_cidrs)
    if len(targeter.fqdns) > 0:
        if table:
            print.Printer.print_fqdns_as_table(analyzer.analyzed_fqdns)
        elif json:
            print.Printer.print_as_json(analyzer.analyzed_fqdns)
        else:
            print.Printer.print_fqdns_as_table(analyzer.analyzed_fqdns)
    if len(targeter.urls) > 0:
        if table:
            print.Printer.print_urls_as_table(analyzer.analyzed_urls)
        elif json:
            print.Printer.print_as_json(analyzer.analyzed_urls)
        else:
            print.Printer.print_urls_as_table(analyzer.analyzed_urls)
    if len(targeter.invalids) > 0:
        if table:
            print.Printer.print_invalids_as_table(targeter.invalids)
        else:
            print.Printer.print_invalids_as_table(targeter.invalids)


if __name__ == '__main__':
    cli()



import click


class CustomOption(click.Option):
    """This wrapper class helps the `click` library to categorize the CLI options."""

    def __init__(self, *args, **kwargs):
        self.category = kwargs.pop('category', None)
        super().__init__(*args, **kwargs)


class CustomCommand(click.Command):
    """This wrapper class instructs the `click` library to list commands in the order that we define their functions."""

    def list_commands(self, ctx: click.Context) -> list[str]:
        """
        List the commands in the order that we define their functions.

        Args:
            ctx (click.Context): The click context.

        Returns:
            list[str]: The list of commands in the order that we define their functions.

        """
        return list(self.commands)

    def format_options(self, ctx: click.Context, formatter: click.HelpFormatter):
        """Format the options by separating them into categories

        Args:
            ctx (click.Context): The click context.
            formatter (HelpFormatter): Format the text-based help page.
        """
        categories = {}

        for param in self.get_params(ctx):
            if isinstance(param, click.Option):
                cat = getattr(param, 'category', 'OTHER')
                categories.setdefault(cat, []).append(param)

        for category, params in categories.items():
            with formatter.section(category):
                formatter.write_dl([(p.opts[0], p.help or '') for p in params])



import pydantic
import typing


class CIDR(pydantic.BaseModel):
    type: str = 'cidr'
    cidr: str = ''
    number_of_hosts: int = 0
    visibility: str = ''
    asn_country_code: str = ''
    asn_description: str = ''
    asn_network: str = ''
    geoip_continent: str = ''
    geoip_country: str = ''


class IPV4(pydantic.BaseModel):
    type: str = 'ipv4'
    ipv4: str = ''
    visibility: str = ''
    asn_country_code: str = ''
    asn_description: str = ''
    asn_network: str = ''
    geoip_continent: str = ''
    geoip_country: str = ''
    pingable: bool = False


class FQDN(pydantic.BaseModel):
    type: str = 'fqdn'
    fqdn: str = ''
    dns_chain: list[str] = ''
    hosts_found: bool = False
    destination_ips: list[IPV4] = []


class URL(pydantic.BaseModel):
    type: str = 'url'
    url: str = ''
    scheme: str = ''
    username: str = ''
    password: str = ''
    port: int = ''
    path: str = ''
    reachable: bool = False
    fqdn: FQDN = None


class GeoIPRecord(pydantic.BaseModel):
    network: str
    geoname_id: int
    continent_code: typing.Optional[str]
    continent_name: typing.Optional[str]
    country_iso_code: typing.Optional[str]
    country_name: typing.Optional[str]
    is_anonymous_proxy: int
    is_satellite_provider: int



import pydantic
import networkx as nx
import matplotlib.pyplot as plt
import numpy as np

import PIL
import pathlib
import os.path

import models
import verbose

BASE_DIR = pathlib.Path(__file__).resolve().parent


class Visualizer(pydantic.BaseModel):
    """Visualizes the data using a network graph."""

    def create_visualization_image(
        self,
        analyzed_ipv4s: list[models.IPV4],
        analyzed_cidrs: list[models.CIDR],
        analyzed_fqdns: list[models.FQDN],
        analyzed_urls: list[models.URL],
        out_path: str = 'network.png',
    ) -> None:
        verbose.debug('Specify an icon per entity.')
        icons_paths = {
            'ipv4': os.path.join(BASE_DIR, 'icons', 'ipv4.png'),
            'cidr': os.path.join(BASE_DIR, 'icons', 'cidr.png'),
            'fqdn': os.path.join(BASE_DIR, 'icons', 'fqdn.png'),
            'url': os.path.join(BASE_DIR, 'icons', 'url.png'),
            '404': os.path.join(BASE_DIR, 'icons', '404.png'),
            'target': os.path.join(BASE_DIR, 'icons', 'target.png'),
        }
        icons = {k: PIL.Image.open(fname) for k, fname in icons_paths.items()}

        verbose.debug('Create the network graph nodes.')
        total = len(analyzed_ipv4s) + len(analyzed_cidrs) + len(analyzed_fqdns) + len(analyzed_urls)
        G = nx.Graph()
        if total > 1:
            G.add_node('center', image=icons['target'], name='center')
        for ip in analyzed_ipv4s:
            G.add_node(ip.ipv4, image=icons['ipv4'], name=ip.ipv4)
            if total > 1:
                G.add_edge('center', ip.ipv4)

        for c in analyzed_cidrs:
            G.add_node(c.cidr, image=icons['cidr'], name=c.cidr)
            if total > 1:
                G.add_edge('center', c.cidr)

        for f in analyzed_fqdns:
            G.add_node(f.fqdn, image=icons['fqdn'], name=f.fqdn)
            if total > 1:
                G.add_edge('center', f.fqdn)

            _chain_wo_fqdn = [node for node in f.dns_chain if node != f.fqdn]
            if _chain_wo_fqdn:
                for link in _chain_wo_fqdn:
                    G.add_node(link, image=icons['fqdn'], name=link)
                    G.add_edge(f.fqdn, link)

            if f.hosts_found:
                for ip in f.destination_ips:
                    _ipnode = f'{f.fqdn}_{ip}'
                    G.add_node(_ipnode, image=icons['ipv4'], name=ip.ipv4)
                    G.add_edge(f.dns_chain[-1], _ipnode)
            else:
                error = f'{f.fqdn} not found'
                G.add_node(error, image=icons['404'], name='404', kind='404')
                G.add_edge(f.dns_chain[-1], error)

        for u in analyzed_urls:
            G.add_node(u.url, image=icons['url'], name=u.url)
            if total > 1:
                G.add_edge('center', u.url)

            G.add_node(u.fqdn.fqdn, image=icons['fqdn'], name=u.fqdn.fqdn)
            G.add_edge(u.url, u.fqdn.fqdn)

            _chain_wo_fqdn = [node for node in u.fqdn.dns_chain if node != u.fqdn.fqdn]
            if _chain_wo_fqdn:
                for link in _chain_wo_fqdn:
                    G.add_node(link, image=icons['fqdn'], name=link)
                    G.add_edge(u.fqdn.fqdn, link)

            if u.fqdn.hosts_found:
                for ip in u.fqdn.destination_ips:
                    _ipnode = f'{u.fqdn.fqdn}_{ip}'
                    G.add_node(_ipnode, image=icons['ipv4'], name=ip.ipv4)
                    G.add_edge(u.fqdn.dns_chain[-1], _ipnode)
            else:
                error = f'{u.fqdn.fqdn} not found'
                G.add_node(error, image=icons['404'], name='404', kind='404')
                G.add_edge(u.fqdn.dns_chain[-1], error)

        verbose.debug('Initialize the figure.')
        pos = nx.kamada_kawai_layout(G)
        if total > 1:
            pos['center'] = np.array([0.0, 0.0])
        fig, ax = plt.subplots(figsize=(19.2, 10.8))

        verbose.debug('Draw the edges.')
        nx.draw_networkx_edges(
            G,
            pos=pos,
            ax=ax,
            arrows=True,
            min_source_margin=15,
            min_target_margin=15,
        )

        # --- Draw ---
        verbose.debug('Draw the icons.')
        tr_figure = ax.transData.transform
        tr_axes = fig.transFigure.inverted().transform
        icon_size = (ax.get_xlim()[1] - ax.get_xlim()[0]) * 0.01
        icon_center = icon_size / 2.0
        for n in G.nodes:
            xf, yf = tr_figure(pos[n])
            xa, ya = tr_axes((xf, yf))
            # get overlapped axes and plot icon
            a = plt.axes([xa - icon_center, ya - icon_center, icon_size, icon_size])
            a.imshow(G.nodes[n]['image'])
            a.axis('off')

        verbose.debug('Add the labels.')
        label_options = {'ec': 'k', 'fc': 'white', 'alpha': 0.7}
        for n in G.nodes:
            node_name = G.nodes[n].get('name', '')

            if node_name == 'center':
                continue

            ax.text(
                pos[n][0],
                pos[n][1] + 0.035,  # offset upward
                node_name,
                fontsize=6,
                bbox=label_options,
                ha='center',  # <-- center horizontally
                va='bottom',  # <-- place box just above the node
            )

        verbose.debug('Save the image.')
        ax.set_axis_off()  # hides axes and ticks
        plt.savefig(out_path, dpi=100, bbox_inches='tight')



import fqdn
import pydantic
import rich.panel

import ipaddress
import sys
import urllib.parse

import verbose


class Targeter(pydantic.BaseModel):
    """Parses the `targets` stores them in different kind of formats."""

    ipv4s: list[str] = []
    ipv4s_with_port: list[str] = []
    ipv6s: list[str] = []
    ipv6s_with_port: list[str] = []
    fqdns: list[str] = []
    fqdns_with_port: list[str] = []
    cidrs_v4: list[str] = []
    cidrs_v6: list[str] = []
    urls: list[str] = []
    invalids: list[str] = []

    def parse_targets_file(self, targets_filepath: str) -> None:
        """
        Parses the targets from a text file or stdin ("-").

        Args:
            targets_filepath (str): Path to file or "-" to read from stdin.

        """
        file = sys.stdin if targets_filepath == '-' else open(targets_filepath, 'r', encoding='utf-8')

        with file as f:
            for line in f:
                val = line.strip()
                if self._validate_ipv4(val):
                    self.ipv4s.append(val)
                elif self._validate_ipv4_with_port(val):
                    self.ipv4s_with_port.append(val)
                elif self._validate_ipv6(val):
                    self.ipv6s.append(val)
                elif self._validate_ipv6_with_port(val):
                    self.ipv6s_with_port.append(val)
                elif self._validate_fqdn(val):
                    self.fqdns.append(val)
                elif self._validate_fqdn_with_port(val):
                    self.fqdns_with_port.append(val)
                elif self._validate_cidr_ipv4(val):
                    self.cidrs_v4.append(val)
                elif self._validate_url(val):
                    self.urls.append(val)
                else:
                    self.invalids.append(val)

        self._remove_duplicates()
        self._sort_ascending()

    def parse_targets_str(self, targets_str: str) -> None:
        """
        Parses the targets from a string.

        Args:
            targets_str (str): The text that contains the targets separted by a newline.

        """
        for val in targets_str.split(','):
            if self._validate_ipv4(val):
                self.ipv4s.append(val)
            elif self._validate_ipv4_with_port(val):
                self.ipv4s_with_port.append(val)
            elif self._validate_ipv6(val):
                self.ipv6s.append(val)
            elif self._validate_ipv6_with_port(val):
                self.ipv6s_with_port.append(val)
            elif self._validate_fqdn(val):
                self.fqdns.append(val)
            elif self._validate_fqdn_with_port(val):
                self.fqdns_with_port.append(val)
            elif self._validate_cidr_ipv4(val):
                self.cidrs_v4.append(val)
            elif self._validate_url(val):
                self.urls.append(val)
            else:
                self.invalids.append(val)

        self._remove_duplicates()
        self._sort_ascending()

    def parse_exclusions_file(self, exclusions_filepath: str) -> None:
        """Parses targets in the exclusions file and removes them from the targets list."""
        # Extracting the exclusions and removing them from the target lists
        with open(exclusions_filepath, 'r', encoding='utf-8') as file:
            for val in file:
                val = val.strip()
                if self._validate_ipv4(val):
                    self._remove_from_list(self.ipv4s, val)
                elif self._validate_ipv4_with_port(val):
                    self._remove_from_list(self.ipv4s_with_port, val)
                elif self._validate_ipv6(val):
                    self._remove_from_list(self.ipv6s, val)
                elif self._validate_ipv6_with_port(val):
                    self._remove_from_list(self.ipv6s_with_port, val)
                elif self._validate_fqdn(val):
                    self._remove_from_list(self.fqdns, val)
                elif self._validate_fqdn_with_port(val):
                    self._remove_from_list(self.fqdns_with_port, val)
                elif self._validate_cidr_ipv4(val):
                    self._remove_from_list(self.cidrs_v4, val)
                elif self._validate_url(val):
                    self._remove_from_list(self.urls, val)
                else:
                    self._remove_from_list(self.invalids, val)

    def parse_exclusions_str(self, exclusion_str: str) -> None:
        for val in exclusion_str.split(','):
            val = val.strip()
            if self._validate_ipv4(val):
                self._remove_from_list(self.ipv4s, val)
            elif self._validate_ipv4_with_port(val):
                self._remove_from_list(self.ipv4s_with_port, val)
            elif self._validate_ipv6(val):
                self._remove_from_list(self.ipv6s, val)
            elif self._validate_ipv6_with_port(val):
                self._remove_from_list(self.ipv6s_with_port, val)
            elif self._validate_fqdn(val):
                self._remove_from_list(self.fqdns, val)
            elif self._validate_fqdn_with_port(val):
                self._remove_from_list(self.fqdns_with_port, val)
            elif self._validate_cidr_ipv4(val):
                self._remove_from_list(self.cidrs_v4, val)
            elif self._validate_url(val):
                self._remove_from_list(self.urls, val)
            else:
                self._remove_from_list(self.invalids, val)

    def total_count(self) -> int:
        return (
            len(self.ipv4s)
            + len(self.ipv4s_with_port)
            + len(self.ipv6s)
            + len(self.ipv6s_with_port)
            + len(self.fqdns)
            + len(self.fqdns_with_port)
            + len(self.cidrs_v4)
            + len(self.cidrs_v6)
            + len(self.urls)
        )

    def print_targets(self) -> None:
        """Prints the parsed targets in a beautifull format."""
        console = verbose.CONSOLE

        if len(self.ipv4s) > 0:
            console.print(rich.panel.Panel.fit('IP Addresses (v4)'))
            for i in self.ipv4s:
                console.print(f' - {i}', highlight=False)

        if len(self.ipv4s_with_port) > 0:
            console.print(rich.panel.Panel.fit('IP Addresses with Port (v4)'))
            for i in self.ipv4s_with_port:
                console.print(f' - {i}', highlight=False)

        if len(self.ipv6s) > 0:
            console.print(rich.panel.Panel.fit('IP Addresses (v6)'))
            for i in self.ipv6s:
                console.print(f' - {i}', highlight=False)

        if len(self.ipv6s_with_port) > 0:
            console.print(rich.panel.Panel.fit('IP Addresses with Port (v6)'))
            for i in self.ipv6s_with_port:
                console.print(f' - {i}', highlight=False)

        if len(self.fqdns) > 0:
            console.print(rich.panel.Panel.fit('FQDNs'))
            for f in self.fqdns:
                console.print(f' - {f}', highlight=False)

        if len(self.fqdns_with_port) > 0:
            console.print(rich.panel.Panel.fit('FQDNs with Port'))
            for f in self.fqdns_with_port:
                console.print(f' - {f}', highlight=False)

        if len(self.cidrs_v4) > 0:
            console.print(rich.panel.Panel.fit('CIDRs (v4)'))
            for c in self.cidrs_v4:
                console.print(f' - {c}', highlight=False)

        if len(self.cidrs_v6) > 0:
            console.print(rich.panel.Panel.fit('CIDRs (v6)'))
            for c in self.cidrs_v6:
                console.print(f' - {c}', highlight=False)

        if len(self.urls) > 0:
            console.print(rich.panel.Panel.fit('URLs'))
            for u in self.urls:
                console.print(f' - {u}', highlight=False)

        if len(self.invalids) > 0:
            console.print(rich.panel.Panel.fit('Invalids'))
            for i in self.invalids:
                console.print(f' - {i}', highlight=False)

    def _remove_from_list(self, target_list: list[str], item: str) -> None:
        """Helper function to remove an item from a list if it exists."""
        if item in target_list:
            target_list.remove(item)

    def _remove_duplicates(self) -> None:
        """Remove all duplicate entries of the targets lists."""
        self.ipv4s = list(set(self.ipv4s))
        self.ipv4s_with_port = list(set(self.ipv4s_with_port))
        self.ipv6s = list(set(self.ipv6s))
        self.ipv6s_with_port = list(set(self.ipv6s_with_port))
        self.fqdns = list(set(self.fqdns))
        self.fqdns_with_port = list(set(self.fqdns_with_port))
        self.cidrs_v4 = list(set(self.cidrs_v4))
        self.urls = list(set(self.urls))
        self.invalids = list(set(self.invalids))

    def _sort_ascending(self) -> None:
        """Sort the targets in ascending order."""
        self.ipv4s.sort()
        self.ipv4s_with_port.sort()
        self.ipv6s.sort()
        self.ipv6s_with_port.sort()
        self.fqdns.sort()
        self.fqdns_with_port.sort()
        self.cidrs_v4.sort()
        self.urls.sort()
        self.invalids.sort()

    def _validate_ipv4(self, value: str) -> bool:
        try:
            ipaddress.IPv4Address(value)
            return True
        except Exception:
            return False

    def _validate_ipv4_with_port(self, value: str) -> bool:
        parts = value.split(':')
        if len(parts) != 2:
            return False

        ipv4, port = parts

        try:
            ipaddress.IPv4Address(ipv4)
            return 1 <= int(port) <= 65535
        except Exception:
            return False

    def _validate_ipv6(self, value: str) -> bool:
        try:
            ipaddress.IPv6Address(value)
            return True
        except Exception:
            return False

    def _validate_ipv6_with_port(self, value: str) -> bool:
        if not (value.startswith('[') and ']:' in value):
            return False

        ipv6, port = value[1:].split(']:', 1)

        try:
            ipaddress.IPv6Address(ipv6)
            return 1 <= int(port) <= 65535
        except Exception:
            return False

    def _validate_cidr_ipv4(self, value: str) -> bool:
        try:
            ipaddress.IPv4Network(value, strict=True)
            return True
        except Exception:
            return False

    def _validate_cidr_ipv6(self, value: str) -> bool:
        try:
            ipaddress.IPv6Network(value, strict=True)
            return True
        except Exception:
            return False

    def _validate_fqdn(self, value: str) -> bool:
        return fqdn.FQDN(value).is_valid

    def _validate_fqdn_with_port(self, value: str) -> bool:
        if ':' not in value:
            return False

        host, port = value.rsplit(':', 1)

        if not host or not self._validate_fqdn(host):
            return False

        return 1 <= int(port) <= 65535

    def _validate_url(self, value: str) -> bool:
        parsed = urllib.parse.urlparse(value)
        return parsed.scheme in ('http', 'https', 'ftp') and bool(parsed.netloc)



import pydantic
import ipwhois
import dns.resolver
import requests
import pandas

import ipaddress
import subprocess
import os
import concurrent.futures
import random
import urllib.parse
import time

import models
import verbose


class Analyzer(pydantic.BaseModel):
    """Analyzes raw targets and populates them with additiona information."""

    analyzed_ipv4s: list[models.IPV4] = []
    analyzed_cidrs: list[models.CIDR] = []
    analyzed_fqdns: list[models.FQDN] = []
    analyzed_urls: list[models.URL] = []
    geoip_records: list[models.GeoIPRecord] = []

    def parse_geoip_data(self, geoip_csv_database_filepath: str):
        data_frame = pandas.read_csv(geoip_csv_database_filepath).fillna('')
        records = data_frame.to_dict(orient='records')

        for r in records:
            self.geoip_records.append(
                models.GeoIPRecord(
                    network=r.get('network'),
                    geoname_id=r.get('geoname_id'),
                    continent_code=r.get('continent_code'),
                    continent_name=r.get('continent_name'),
                    country_iso_code=r.get('country_iso_code'),
                    country_name=r.get('country_name'),
                    is_anonymous_proxy=r.get('is_anonymous_proxy'),
                    is_satellite_provider=r.get('is_satellite_provider'),
                )
            )

    def analyze_ipv4s(self, ipv4s: list[str], no_threads: int) -> None:
        """Analyze the IP addresses (v4) and populate them with information.

        Args:
            ipv4s (list[str]): A list of IP addresses in raw format.
            no_threads(int): The number of worker threads to run in-parallel.
        """
        verbose.debug('Analyze IPV4s.')

        with concurrent.futures.ThreadPoolExecutor(max_workers=no_threads) as executor:
            futures = [executor.submit(self._populate_ipv4, ip) for ip in ipv4s]

            for future in concurrent.futures.as_completed(futures):
                ipv4_obj = future.result()
                verbose.normal(ipv4_obj.ipv4)
                self.analyzed_ipv4s.append(ipv4_obj)

    def analyze_cidrs(self, cidrs: list[str], no_threads: int) -> None:
        """Analyze the CIDR IP addresses (v4) and populate them with information.

        Args:
            cidrs (list[str]): A list of CIDR IP addresses in raw format.
            no_threads(int): The number of worker threads to run in-parallel.
        """
        verbose.debug('Analyze CIDRs.')

        with concurrent.futures.ThreadPoolExecutor(max_workers=no_threads) as executor:
            futures = [executor.submit(self._populate_cidr, cidr) for cidr in cidrs]

            for future in concurrent.futures.as_completed(futures):
                cidr_obj = future.result()
                verbose.normal(cidr_obj.cidr)
                self.analyzed_cidrs.append(cidr_obj)

    def analyze_fqdns(self, fqdns: list[str], no_threads: int) -> None:
        """Analyze the FQDNs and populate them with information.

        Args:
            fqdns (list[str]): A list of FQDNs in raw format.
            no_threads(int): The number of worker threads to run in-parallel.
        """
        verbose.debug('Analyze FQDNs.')

        with concurrent.futures.ThreadPoolExecutor(max_workers=no_threads) as executor:
            futures = [executor.submit(self._populate_fqdn, fqdn) for fqdn in fqdns]

            for future in concurrent.futures.as_completed(futures):
                fqdn_obj = future.result()
                verbose.normal(fqdn_obj.fqdn)
                self.analyzed_fqdns.append(fqdn_obj)

    def analyze_urls(self, urls: list[str], no_threads: int) -> None:
        """Analyze the URLS and populate them with information.

        Args:
            urls (list[str]): A list of URLs in raw format.
            no_threads(int): The number of worker threads to run in-parallel.
        """
        verbose.debug('Analyze URLs.')

        with concurrent.futures.ThreadPoolExecutor(max_workers=no_threads) as executor:
            futures = [executor.submit(self._populate_url, url) for url in urls]

            for future in concurrent.futures.as_completed(futures):
                url_obj = future.result()
                verbose.normal(url_obj.url)
                self.analyzed_urls.append(url_obj)

    def _populate_ipv4(self, ipv4: str) -> models.IPV4:
        verbose.debug(f'Analyze {ipv4}.')

        ipv4_obj = models.IPV4()
        ipv4_obj.ipv4 = ipv4

        ########
        # Math #
        ########
        ip = ipaddress.IPv4Address(ipv4)

        ##############
        # Visibility #
        ##############
        ipv4_obj.visibility = 'Private' if ip.is_private else 'Public'

        ########
        # RDAP #
        ########
        if ipv4_obj.visibility == 'Public':
            while True:
                try:
                    whois = ipwhois.IPWhois(ipv4_obj.ipv4)
                    rdap = whois.lookup_rdap(depth=1) or {}
                    break
                except Exception:
                    time.sleep(5)
                    continue
            ipv4_obj.asn_network = rdap.get('network', {}).get('name', '').replace(',', '')
            ipv4_obj.asn_country_code = rdap.get('asn_country_code')
            ipv4_obj.asn_description = rdap.get('asn_description', '').replace(',', '')
        else:
            ipv4_obj.asn_network = 'N/A'
            ipv4_obj.asn_country_code = 'N/A'
            ipv4_obj.asn_description = 'N/A'

        #########
        # GeoIP #
        #########
        if ipv4_obj.visibility == 'Public':
            for r in self.geoip_records:
                network = ipaddress.IPv4Network(r.network)

                if ip in network:
                    ipv4_obj.geoip_continent = r.continent_name
                    ipv4_obj.geoip_country = r.country_name
        else:
            ipv4_obj.geoip_continent = 'N/A'
            ipv4_obj.geoip_country = 'N/A'

        ########
        # Ping #
        ########
        if ipv4_obj.visibility == 'Public':
            param = '-n' if os.sys.platform.lower() == 'win32' else '-c'
            command = ['ping', param, '1', '-i', '0.2', ipv4]
            pingable = subprocess.call(command, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL) == 0
            ipv4_obj.pingable = pingable
        else:
            ipv4_obj.pingable = False

        return ipv4_obj

    def _populate_cidr(self, cidr: str) -> models.CIDR:
        verbose.debug(f'Analyze {cidr}.')

        cidr_obj = models.CIDR()
        cidr_obj.cidr = cidr

        ########
        # Math #
        ########
        network = ipaddress.IPv4Network(cidr, strict=False)
        cidr_obj.number_of_hosts = network.num_addresses - 2

        ##############
        # Visibility #
        ##############
        cidr_obj.visibility = 'Private' if network.is_private else 'Public'

        ########
        # RDAP #
        ########
        if cidr_obj.visibility == 'Public':
            ip = cidr.split('/')[0]
            while True:
                try:
                    whois = ipwhois.IPWhois(ip)
                    rdap = whois.lookup_rdap(depth=1) or {}
                    break
                except Exception:
                    time.sleep(5)
                    continue
            cidr_obj.asn_network = rdap.get('network', {}).get('name', '').replace(',', '')
            cidr_obj.asn_country_code = rdap.get('asn_country_code')
            cidr_obj.asn_description = rdap.get('asn_description', '').replace(',', '')
        else:
            cidr_obj.asn_network = 'N/A'
            cidr_obj.asn_country_code = 'N/A'
            cidr_obj.asn_description = 'N/A'

        #########
        # GeoIP #
        #########
        if cidr_obj.visibility == 'Public':
            for r in self.geoip_records:
                if cidr_obj.asn_network == r.network:
                    cidr_obj.geoip_continent = r.continent_name
                    cidr_obj.geoip_country = r.country_name
        else:
            cidr_obj.geoip_continent = 'N/A'
            cidr_obj.geoip_country = 'N/A'

        return cidr_obj

    def _populate_fqdn(self, fqdn: str) -> models.FQDN:
        verbose.debug(f'Analyze {fqdn}.')

        ######################################################################################
        # Process                                                                            #
        # ---                                                                                #
        # Notes:                                                                             #
        # - We need to use different DNS every time otherwise we trigger a DOS response.     #
        # - A FQDN can point only to a *single* CNAME (RFC 1034 & RFC 2181).                 #
        # - A FQDN can point to multiple IP addresses (Load-Balancing).                      #
        # - A FQDN cannot have a CNAME and A/AAA records at the same time.                   #
        # Code:                                                                              #
        # - Keep a list of FQDNs to resolve.                                                 #
        # - Ask DNS server for CNAME records of FQDN.                                        #
        # - If there is a CNAME record, append it in the list of FQDNs to resolve.           #
        #    - else check for A records for the hostname.                                    #
        ######################################################################################
        f = models.FQDN(fqdn=fqdn, dns_chain=[fqdn], destination_ips=[])

        #############################
        # Discover the CNAME Chain. #
        #############################
        while True:
            cname_record = ''
            try:
                answer = dns.resolver.resolve_at(random.choice(DNS_SERVERS), f.dns_chain[-1], 'CNAME')
                time.sleep(1)

                for rdap in answer:
                    cname_record = str(rdap.target).rstrip('.')  # Remove the trailing dot.

            except dns.resolver.NXDOMAIN:
                # NXDOMAIN stands for Non-Existent Domain.
                break

            except dns.resolver.NoAnswer:
                # The domain does exist, but the specific DNS record type you're asking for is missing.
                break

            except dns.resolver.LifetimeTimeout:
                # The resolution lifetime expired.
                verbose.debug('dns.resolver.LifetimeTimeout')
                continue

            except dns.resolver.NoNameservers:
                # If no non-broken nameservers are available to answer the question.
                verbose.debug('dns.resolver.NoNameservers')
                continue

            f.dns_chain.append(cname_record)

        ############################################################
        # For the last link in the DNS chain, check its A records. #
        ############################################################
        while True:
            try:
                answer = dns.resolver.resolve_at(random.choice(DNS_SERVERS), f.dns_chain[-1], 'A')
                time.sleep(1)

                resolved_ips = []
                for rdap in answer:
                    resolved_ips.append(str(rdap.address))

                f.hosts_found = True

                for ip in resolved_ips:
                    ip_obj = self._populate_ipv4(ip)
                    f.destination_ips.append(ip_obj)

                break

            except dns.resolver.NXDOMAIN:
                # NXDOMAIN stands for Non-Existent Domain.
                break

            except dns.resolver.NoAnswer:
                # The domain does exist, but the specific DNS record type you're asking for is missing.
                break

            except dns.resolver.LifetimeTimeout:
                # The resolution lifetime expired.
                verbose.debug('dns.resolver.LifetimeTimeout')
                continue

            except dns.resolver.NoNameservers:
                # If no non-broken nameservers are available to answer the question.
                verbose.debug('dns.resolver.NoNameservers')
                continue

        return f

    def _populate_url(self, url: str) -> models.URL:
        verbose.debug(f'Analyze {url}.')

        parsed_url = urllib.parse.urlparse(url)
        parsed_port = parsed_url.port if parsed_url.port is not None else 443 if parsed_url.scheme == 'https' else 80

        u = models.URL(
            url=url,
            scheme=parsed_url.scheme,
            username=parsed_url.username if parsed_url.username is not None else '',
            password=parsed_url.password if parsed_url.password is not None else '',
            port=parsed_port,
            path=parsed_url.path,
        )

        u.fqdn = self._populate_fqdn(parsed_url.hostname)

        ########
        # CURL #
        ########
        try:
            requests.get(u.url, verify=False, timeout=2)
            u.reachable = True
        except requests.exceptions.RequestException:
            u.reachable = False

        return u


DNS_SERVERS = [
    # Google
    '8.8.8.8',
    # Google
    '8.8.4.4',
    # Control D
    '76.76.2.0',
    # Control D
    '76.76.10.0',
    # Quad9
    '9.9.9.9',
    # Quad9
    '149.112.112.112',
    # OpenDNS Home
    '208.67.222.222',
    # OpenDNS Home
    '208.67.220.220',
    # Cloudflare
    '1.1.1.1',
    # Cloudflare
    '1.0.0.1',
    # AdGuard DNS
    '94.140.14.14',
    # AdGuard DNS
    '94.140.15.15',
    # CleanBrowsing
    '185.228.168.9',
    # CleanBrowsing
    '185.228.169.9',
    # Oracle
    '216.146.35.35',
    # Oracle
    '216.146.36.36',
    # Quad101
    '101.101.101.101',
    # Quad101
    '101.102.103.104',
    # Nippon Telegraph and Telephone
    '129.250.35.250',
    # Nippon Telegraph and Telephone
    '129.250.35.251',
]



import rich.table
import rich.json

import models
import verbose


class Printer:
    """Prints the target's analysis' results in various formats."""

    #
    # ALL
    #

    @staticmethod
    def print_as_json(targets: list[any]) -> None:
        for t in targets:
            verbose.normal(rich.json.JSON(t.model_dump_json(), indent=None))

    #
    # CIDRs
    #

    @staticmethod
    def print_cidrs_as_table(cidrs: list[models.CIDR]) -> None:
        t = rich.table.Table(box=rich.box.ASCII)

        t.add_column('CIDR')
        t.add_column('Possible # of Hosts')
        t.add_column('Visibility')
        t.add_column('ASN Country')
        t.add_column('ASN Description')
        t.add_column('ASN Network')
        t.add_column('GeoIP Continent')
        t.add_column('GeoIP Country')

        for cidr in cidrs:
            t.add_row(
                cidr.cidr,
                str(cidr.number_of_hosts),
                cidr.visibility,
                cidr.asn_country_code,
                cidr.asn_description,
                cidr.asn_network,
                cidr.geoip_continent,
                cidr.geoip_country,
            )

        verbose.normal(t)

    @staticmethod
    def print_cidrs_as_raw(cidrs: list[models.CIDR]) -> None:
        for cidr in cidrs:
            verbose.normal(
                f'[white]{cidr.type}[/white],[green]{cidr.cidr}[/green],[yellow]{cidr.number_of_hosts}[/yellow],[yellow]{cidr.visibility}[/yellow],[red]{cidr.asn_country_code}[/red],[red]{cidr.asn_description}[/red],[red]{cidr.asn_network}[/red]',
            )

    #
    # IPs
    #

    @staticmethod
    def print_ipv4s_as_table(ipv4s: list[models.IPV4]) -> None:
        t = rich.table.Table(box=rich.box.ASCII)

        t.add_column('IP Address (v4)')
        t.add_column('Visibility')
        t.add_column('ASN Country')
        t.add_column('ASN Description')
        t.add_column('ASN Network')
        t.add_column('GeoIP Continent')
        t.add_column('GeoIP Country')
        t.add_column('Pingable')

        for ipv4 in ipv4s:
            t.add_row(
                ipv4.ipv4,
                ipv4.visibility,
                ipv4.asn_country_code,
                ipv4.asn_description,
                ipv4.asn_network,
                ipv4.geoip_continent,
                ipv4.geoip_country,
                'yes' if ipv4.pingable else 'no',
            )

        verbose.normal(t)

    def print_ipv4s_as_raw(ipv4s: list[models.IPV4]) -> None:
        for ipv4 in ipv4s:
            verbose.normal(
                f'[white]{ipv4.type}[/white],'
                f'[green]{ipv4.ipv4}[/green],'
                f'[yellow]{ipv4.visibility}[/yellow],'
                f'[red]{ipv4.asn_country_code}[/red],'
                f'[red]{ipv4.asn_description}[/red],'
                f'[red]{ipv4.asn_network}[/red],'
                f'[blue]{"pingable" if ipv4.pingable else "not pingable"}[/blue]'
            )

    #
    # FQDNs
    #

    @staticmethod
    def print_fqdns_as_table(fqdns: list[models.FQDN]) -> None:
        t = rich.table.Table(box=rich.box.ASCII)

        t.add_column('FQDN')
        t.add_column('DNS Chain')
        t.add_column('ASN Country')
        t.add_column('ASN Description')
        t.add_column('ASN Network')
        t.add_column('GeoIP Continent')
        t.add_column('GeoIP Country')
        t.add_column('Pingable')

        for fqdn in fqdns:
            if fqdn.hosts_found:
                for ip in fqdn.destination_ips:
                    t.add_row(
                        fqdn.fqdn,
                        ' > '.join(fqdn.dns_chain) + f' > {ip.ipv4}',
                        ip.asn_country_code,
                        ip.asn_description,
                        ip.asn_network,
                        ip.geoip_continent,
                        ip.geoip_country,
                        'yes' if ip.pingable else 'no',
                    )
            else:
                t.add_row(
                    fqdn.fqdn,
                    ' > '.join(fqdn.dns_chain) + ' > Not Found',
                    'N/A',
                    'N/A',
                    'N/A',
                    'N/A',
                    'N/A',
                    'N/A',
                )

        verbose.normal(t)

    @staticmethod
    def print_fqdns_as_raw(fqdns: list[models.FQDN]) -> None:
        for fqdn in fqdns:
            if fqdn.hosts_found:
                for ip in fqdn.destination_ips:
                    verbose.normal(
                        f'[white]{fqdn.type}[/white],[green]{fqdn.fqdn}[/green],[yellow]{" > ".join(fqdn.dns_chain) + f" > {ip}"}[/yellow],[red]{ip.asn_country_code}[/red],[red]{ip.asn_description}[/red],[red]{ip.asn_network}[/red],[red]{ip.geoip_continent}[/red],[red]{ip.geoip_country}[/red],[blue]{"yes" if ip.pingable else "no"}[/blue]',
                    )
            else:
                verbose.normal(
                    f'[white]{fqdn.type}[/white],[green]{fqdn.fqdn}[/green],[yellow]{" > ".join(fqdn.dns_chain) + " > Not Found"}[/yellow],[red]N/A[/red],[red]N/A[/red],[red]N/A[/red],[red]N/A[/red],[red]N/A[/red],[blue]N/A[/blue]',
                )

    #
    # URLs
    #

    @staticmethod
    def print_urls_as_table(urls: list[models.URL]) -> None:
        t = rich.table.Table(box=rich.box.ASCII)

        t.add_column('URL')
        t.add_column('DNS Chain')
        t.add_column('ASN Country')
        t.add_column('ASN Description')
        t.add_column('ASN Network')
        t.add_column('GeoIP Continent')
        t.add_column('GeoIP Country')
        t.add_column('Pingable')
        t.add_column('Reachable')

        for url in urls:
            if url.fqdn.hosts_found:
                for ip in url.fqdn.destination_ips:
                    t.add_row(
                        url.url,
                        ' > '.join(url.fqdn.dns_chain) + f' > {ip.ipv4}',
                        ip.asn_country_code,
                        ip.asn_description,
                        ip.asn_network,
                        ip.geoip_continent,
                        ip.geoip_country,
                        'yes' if ip.pingable else 'no',
                        'yes' if url.reachable else 'no',
                    )
            else:
                t.add_row(
                    url.url,
                    ' > '.join(url.fqdn.dns_chain) + ' > Not Found',
                    'N/A',
                    'N/A',
                    'N/A',
                    'N/A',
                    'N/A',
                    'N/A',
                    'N/A',
                )

        verbose.normal(t)

    @staticmethod
    def print_urls_as_raw(urls: list[models.URL]) -> None:
        for url in urls:
            if url.fqdn.hosts_found:
                for ip in url.fqdn.destination_ips:
                    verbose.normal(
                        f'[white]{url.type}[/white],[green]{url.url}[/green],[yellow]{" > ".join(url.fqdn.dns_chain) + f" > {ip}"}[/yellow],[red]{ip.asn_country_code}[/red],[red]{ip.asn_description}[/red],[red]{ip.asn_network}[/red],[red]{ip.geoip_continent}[/red],[red]{ip.geoip_country}[/red],[blue]{"yes" if ip.pingable else "no"}[/blue],[blue]{"yes" if url.reachable else "no"}[/blue]',
                    )
            else:
                verbose.normal(
                    f'[white]{url.type}[/white],[green]{url.url}[/green],[yellow]{" > ".join(url.fqdn.dns_chain) + " > Not Found"}[/yellow],[red]N/A[/red],[red]N/A[/red],[red]N/A[/red],[red]N/A[/red],[red]N/A[/red],[blue]N/A[/blue],[blue]N/A[/blue]',
                )

    #
    # Invalids
    #

    @staticmethod
    def print_invalids_as_table(invalids: list[str]) -> None:
        t = rich.table.Table(box=rich.box.ASCII)

        t.add_column('Invalid')

        for invalid in invalids:
            t.add_row(invalid)

        verbose.normal(t)

    @staticmethod
    def print_invalids_as_raw(invalids: list[str]) -> None:
        for invalid in invalids:
            verbose.normal(f'[white]invalid[/white],[red]{invalid}[/red]')



import click

import os
import hashlib


def validate_file_exists(ctx, param, value):
    if not _file_exists(value) and value != '':
        raise click.BadParameter(f"the file '{value}' does not exist.")

    return value


def validate_png_filename(ctx, param, value):
    if value:
        if value != '' and not value.lower().endswith('.png'):
            raise click.BadParameter("Output filename must end with '.png'.")
    return value


def _file_exists(filepath: str) -> bool:
    """
    Checks if a file exists at the given path.

    Args:
        filepath (str): The path to the file.

    Returns:
        bool: `True` if the file exists, otherwise `False`.

    """
    return os.path.exists(filepath)


def _verify_sha256(filepath: str, expected_hash: str) -> bool:
    sha256 = hashlib.sha256()

    with open(filepath, 'rb') as f:
        for chunk in iter(lambda: f.read(8192), b''):
            sha256.update(chunk)

    file_hash = sha256.hexdigest()

    return file_hash == expected_hash.lower()



from __version__ import __version__

SILENT = False
HIGHLIGHT = False
SOFT_WRAP = False
DEBUG = False
CONSOLE = {}


def print_banner(silent: bool, highlight: bool = True) -> None:
    logo = f"""
 ▗▄▄▖ ▗▄▄▖ ▗▄▖ ▗▄▄▖ ▗▄▄▄▖▗▄▄▄▄▖
▐▌   ▐▌   ▐▌ ▐▌▐▌ ▐▌▐▌      ▗▞▘
 ▝▀▚▖▐▌   ▐▌ ▐▌▐▛▀▘ ▐▛▀▀▘ ▗▞▘  
▗▄▄▞▘▝▚▄▄▖▝▚▄▞▘▐▌   ▐▙▄▄▖▐▙▄▄▄▖ {__version__}
        
        https://github.com/xcalts/scopez
"""
    if not silent:
        CONSOLE.print(logo, highlight=False)


def critical(message: str) -> None:
    if not SILENT:
        CONSOLE.print(f'[bold red][CRITICAL][/bold red] {message}', highlight=HIGHLIGHT, soft_wrap=SOFT_WRAP)


def error(message: str) -> None:
    if not SILENT:
        CONSOLE.print(f'[red][ERROR]   [/red] {message}', highlight=HIGHLIGHT, soft_wrap=SOFT_WRAP)


def warning(message: str) -> None:
    if not SILENT:
        CONSOLE.print(f'[yellow][WARNING] [/yellow] {message}', highlight=HIGHLIGHT, soft_wrap=SOFT_WRAP)


def info(message: str) -> None:
    if not SILENT:
        CONSOLE.print(f'[green][INFO]    [/green] {message}', highlight=HIGHLIGHT, soft_wrap=SOFT_WRAP)


def debug(message: str) -> None:
    if not SILENT and DEBUG:
        CONSOLE.print(f'[blue][DEBUG]   [/blue] {message}', highlight=HIGHLIGHT, soft_wrap=SOFT_WRAP)


def normal(message: str) -> None:
    if not SILENT:
        CONSOLE.print(message, highlight=HIGHLIGHT, soft_wrap=SOFT_WRAP)
